@@ -34,6 +34,7 @@ import socket
 import sys
 import time
 from statistics import stdev
+from typing import Any, Tuple, Optional
 
 import dns.edns
 import dns.exception
@@ -48,8 +49,8 @@ import dns.resolver
 import httpx
 
 from dnsdiag.dns import PROTO_UDP, PROTO_TCP, PROTO_TLS, PROTO_HTTPS, PROTO_QUIC, PROTO_HTTP3, proto_to_text, \
-    unsupported_feature, random_string, getDefaultPort, valid_rdatatype
-from dnsdiag.shared import __version__
+    get_default_port, valid_rdatatype
+from dnsdiag.shared import __version__, valid_hostname, unsupported_feature, random_string, die, err
 
 __author__ = 'Babak Farrokhi (babak@farrokhi.net)'
 __license__ = 'BSD'
@@ -57,7 +58,7 @@ __progname__ = os.path.basename(sys.argv[0])
 shutdown = False
 
 
-def usage():
+def usage(exit_code: int = 0) -> None:
     print("""%s version %s
 Usage: %s [-346aDeEFhLmqnrvTQxXH] [-i interval] [-w wait] [-p dst_port] [-P src_port] [-S src_ip]
        %s [-c count] [-t qtype] [-C class] [-s server] [--ecs client_subnet] hostname
@@ -88,37 +89,32 @@ Usage: %s [-346aDeEFhLmqnrvTQxXH] [-i interval] [-w wait] [-p dst_port] [-P src_
   -e, --edns        Enable EDNS0 and set its options
   -E, --ede         (Ignored - EDE messages now always displayed when present)
   -n, --nsid        Enable the NSID bit to retrieve resolver identification (implies EDNS)
-      --cookie      Display EDNS cookies when present
+      --cookie      Display EDNS cookies when present (implies EDNS)
   -D, --dnssec      Enable the DNSSEC desired flag (implies EDNS)
-      --ecs         Set EDNS Client Subnet option (format: IP/prefix, e.g., 192.168.1.0/24)
+      --ecs         Set EDNS Client Subnet option (format: IP/prefix, e.g., 192.168.1.0/24) (implies EDNS)
   -F, --flags       Display response flags
   -x, --expert      Display additional information (implies --ttl, --flags)
 """ % (__progname__, __version__, __progname__, ' ' * len(__progname__)))
-    sys.exit(0)
+    sys.exit(exit_code)
 
 
-def setup_signal_handler():
+def setup_signal_handler() -> None:
     try:
-        signal.signal(signal.SIGTSTP, signal.SIG_IGN)  # ignore CTRL+Z
+        if hasattr(signal, 'SIGTSTP'):
+            signal.signal(signal.SIGTSTP, signal.SIG_IGN)  # ignore CTRL+Z
         signal.signal(signal.SIGINT, signal_handler)  # custom CTRL+C handler
     except AttributeError:  # not all signals are supported on all platforms
         pass
 
 
-def signal_handler(sig, frame):
+def signal_handler(sig: int, frame: Any) -> None:
     global shutdown
     if shutdown:  # pressed twice, so exit immediately
         sys.exit(0)
     shutdown = True  # pressed once, exit gracefully
 
 
-def print_stderr(s, should_die):
-    print(s, file=sys.stderr, flush=True)
-    if should_die:
-        sys.exit(1)
-
-
-def validate_server_address(dnsserver, address_family):
+def validate_server_address(dnsserver: str, address_family: Optional[int]) -> Tuple[str, str]:
     """checks if we have a valid dns server address and resolve if it is a hostname"""
 
     original_server = dnsserver
@@ -126,33 +122,56 @@ def validate_server_address(dnsserver, address_family):
         ipaddress.ip_address(dnsserver)
     except ValueError:  # so it is not a valid IPv4 or IPv6 address, so try to resolve host name
         try:
-            dnsserver = socket.getaddrinfo(dnsserver, port=None, family=address_family)[1][4][0]
-        except OSError:
-            print_stderr('Error: cannot resolve hostname: %s' % dnsserver, True)
+            if address_family == socket.AF_INET6:
+                results = socket.getaddrinfo(dnsserver, None, address_family, socket.SOCK_DGRAM,
+                                             flags=socket.AI_V4MAPPED)
+            else:
+                results = socket.getaddrinfo(dnsserver, None, address_family or socket.AF_UNSPEC, socket.SOCK_DGRAM)
+
+            if not results:
+                die(f'ERROR: cannot resolve hostname: {dnsserver}')
+
+            # getaddrinfo returns list of tuples: (family, type, proto, canonname, sockaddr)
+            # Extract IP address from first result's sockaddr tuple
+            family, socktype, proto, canonname, sockaddr = results[0]
+
+            # sockaddr format depends on address family:
+            # IPv4: (address, port)
+            # IPv6: (address, port, flow info, scope id)
+            if sockaddr and len(sockaddr) >= 1:
+                dnsserver = str(sockaddr[0])
+            else:
+                die(f'ERROR: invalid address data for hostname: {dnsserver}')
+
+        except (OSError, socket.gaierror) as e:
+            die(f'ERROR: cannot resolve hostname {original_server}: {e}')
+        except (IndexError, TypeError, ValueError) as e:
+            die(f'ERROR: invalid address format for hostname {original_server}: {e}')
+
     return dnsserver, original_server
 
 
-def main():
+def main() -> None:
     global shutdown
     setup_signal_handler()
 
     if len(sys.argv) == 1:
         usage()
 
-    dns.rdata.load_all_types()
+    dns.rdata.load_all_types()  # type: ignore[no-untyped-call]
     # defaults
     rdatatype = 'A'
     rdata_class = dns.rdataclass.from_text('IN')
     count = 10
     timeout = 2
-    interval = 1
+    interval = 1.0
     quiet = False
     verbose = False
     show_flags = False
     show_cookie = False
     dnsserver = None  # do not try to use system resolver by default
     proto = PROTO_UDP
-    dst_port = getDefaultPort(proto)
+    dst_port = get_default_port(proto)
     use_default_dst_port = True
     src_port = 0
     src_ip = None
@@ -164,7 +183,9 @@ def main():
     force_miss = False
     show_answer = False
     request_flags = dns.flags.from_text('RD')
-    af = socket.AF_INET
+    af = None
+    af_ipv4_set = False
+    af_ipv6_set = False
     qname = 'wikipedia.org'
 
     try:
@@ -173,15 +194,16 @@ def main():
                                     "port=", "srcip=", "tcp", "ipv4", "ipv6", "cache-miss", "srcport=", "edns",
                                     "dnssec", "flags", "norecurse", "tls", "doh", "nsid", "ede", "class=", "ttl",
                                     "expert", "answer", "quic", "http3", "ecs=", "cookie"])
-    except getopt.GetoptError as err:
-        # print help information and exit:
-        print_stderr(err, False)  # will print something like "option -a not recognized"
-        usage()
+    except getopt.GetoptError as getopt_err:
+        err(str(getopt_err))
+        usage(1)
 
     if args and len(args) == 1:
         qname = args[0]
+        if not valid_hostname(qname, allow_underscore=True):
+            die(f"ERROR: invalid hostname: {qname}")
     else:
-        usage()
+        usage(1)
 
     for o, a in opts:
         if o in ("-h", "--help"):
@@ -191,7 +213,7 @@ def main():
             if a.isdigit():
                 count = abs(int(a))
             else:
-                print_stderr("Invalid count of requests: %s" % a, True)
+                die(f"ERROR: invalid count of requests: {a}")
 
         elif o in ("-v", "--verbose"):
             verbose = True
@@ -204,7 +226,12 @@ def main():
             verbose = False
 
         elif o in ("-w", "--wait"):
-            timeout = int(a)
+            try:
+                timeout = int(a)
+                if timeout < 0:
+                    die(f"ERROR: wait time must be non-negative: {a}")
+            except ValueError:
+                die(f"ERROR: invalid wait time value: {a}")
 
         elif o in ("-a", "--answer"):
             show_answer = True
@@ -217,7 +244,12 @@ def main():
             force_miss = True
 
         elif o in ("-i", "--interval"):
-            interval = float(a)
+            try:
+                interval = float(a)
+                if interval < 0:
+                    die(f"ERROR: interval must be non-negative: {a}")
+            except ValueError:
+                die(f"ERROR: invalid interval value: {a}")
 
         elif o in ("-L", "--ttl"):
             show_ttl = True
@@ -229,38 +261,44 @@ def main():
             try:
                 rdata_class = dns.rdataclass.from_text(a)
             except dns.rdataclass.UnknownRdataclass:
-                print_stderr("Invalid RR class: %s" % a, True)
+                die(f"ERROR: invalid RR class: {a}")
 
         elif o in ("-T", "--tcp"):
             proto = PROTO_TCP
             if use_default_dst_port:
-                dst_port = getDefaultPort(proto)
+                dst_port = get_default_port(proto)
 
         elif o in ("-X", "--tls"):
             proto = PROTO_TLS
             if use_default_dst_port:
-                dst_port = getDefaultPort(proto)
+                dst_port = get_default_port(proto)
 
         elif o in ("-H", "--doh"):
             proto = PROTO_HTTPS
             if use_default_dst_port:
-                dst_port = getDefaultPort(proto)
+                dst_port = get_default_port(proto)
 
         elif o in ("-Q", "--quic"):
             proto = PROTO_QUIC
             if use_default_dst_port:
-                dst_port = getDefaultPort(proto)
+                dst_port = get_default_port(proto)
 
         elif o in ("-3", "--http3"):
             proto = PROTO_HTTP3
             if use_default_dst_port:
-                dst_port = getDefaultPort(proto)
+                dst_port = get_default_port(proto)
 
         elif o in ("-4", "--ipv4"):
+            if af_ipv6_set:
+                die("ERROR: cannot specify both -4 and -6")
             af = socket.AF_INET
+            af_ipv4_set = True
 
         elif o in ("-6", "--ipv6"):
+            if af_ipv4_set:
+                die("ERROR: cannot specify both -4 and -6")
             af = socket.AF_INET6
+            af_ipv6_set = True
 
         elif o in ("-e", "--edns"):
             use_edns = True
@@ -284,31 +322,46 @@ def main():
             pass
 
         elif o == "--cookie":
+            use_edns = True  # required
             show_cookie = True
 
         elif o in ("-p", "--port"):
-            dst_port = int(a)
-            use_default_dst_port = False
+            try:
+                dst_port = int(a)
+                if not (0 < dst_port <= 65535):
+                    die(f"ERROR: port must be between 1 and 65535: {a}")
+                use_default_dst_port = False
+            except ValueError:
+                die(f"ERROR: invalid port value: {a}")
 
         elif o in ("-P", "--srcport"):
-            src_port = int(a)
-            if src_port < 1024 and not quiet:
-                print_stderr("WARNING: Source ports below 1024 are only available to superuser", False)
+            try:
+                src_port = int(a)
+                if not (0 <= src_port <= 65535):
+                    die(f"ERROR: source port must be between 0 and 65535: {a}")
+                if src_port < 1024 and not quiet:
+                    err("WARNING: Source ports below 1024 are only available to superuser")
+            except ValueError:
+                die(f"ERROR: invalid source port value: {a}")
 
         elif o in ("-S", "--srcip"):
-            src_ip = a
+            try:
+                ipaddress.ip_address(a)
+                src_ip = a
+            except ValueError:
+                die(f"ERROR: invalid source IP address: {a}")
 
         elif o == "--ecs":
             client_subnet = a
             use_edns = True  # ECS requires EDNS
 
         else:
-            usage()
+            usage(1)
 
     # Use system DNS server if parameter is not specified
     # remember not all systems have /etc/resolv.conf (i.e. Android)
     if dnsserver is None:
-        dnsserver = dns.resolver.get_default_resolver().nameservers[0]
+        dnsserver = str(dns.resolver.get_default_resolver().nameservers[0])
 
     dnsserver_ip, dnsserver_hostname = validate_server_address(dnsserver, af)
 
@@ -317,10 +370,13 @@ def main():
 
     # validate RR type
     if not valid_rdatatype(rdatatype):
-        print_stderr('Error: Invalid record type: %s ' % rdatatype, True)
+        die(f'ERROR: invalid record type: {rdatatype}')
 
     # Display the hostname if it differs from the resolved IP, otherwise just the IP
     server_display = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else dnsserver_ip
+    # Wrap IPv6 addresses in brackets for better readability
+    if ':' in server_display and not server_display.startswith('['):
+        server_display = f"[{server_display}]"
     print("%s DNS: %s:%d, hostname: %s, proto: %s, class: %s, type: %s, flags: [%s]" %
           (__progname__, server_display, dst_port, qname, proto_to_text(proto), dns.rdataclass.to_text(rdata_class),
            rdatatype, dns.flags.to_text(request_flags)), flush=True)
@@ -338,7 +394,7 @@ def main():
             fqdn = qname
 
         if use_edns:
-            edns_options = []
+            edns_options: list[Any] = []
             if want_nsid:
                 edns_options.append(dns.edns.GenericOption(dns.edns.NSID, ''))
             if client_subnet:
@@ -346,12 +402,11 @@ def main():
                     ecs_option = dns.edns.ECSOption.from_text(client_subnet)
                     edns_options.append(ecs_option)
                 except Exception as e:
-                    print_stderr("Error: Invalid ECS format '%s': %s" % (client_subnet, e), True)
+                    die(f"ERROR: invalid ECS format '{client_subnet}': {e}")
             if show_cookie:
                 # Send a client cookie (8 random bytes as per RFC 7873)
-                import os
                 client_cookie = os.urandom(8)
-                edns_options.append(dns.edns.GenericOption(10, client_cookie))  # COOKIE = 10
+                edns_options.append(dns.edns.CookieOption(client_cookie, b''))
 
             query = dns.message.make_query(fqdn, rdatatype, rdata_class, flags=request_flags,
                                            use_edns=True, want_dnssec=want_dnssec, payload=1232,
@@ -387,14 +442,17 @@ def main():
                         else:
                             https_server = dnsserver_ip
                         answers = dns.query.https(query, https_server, timeout=timeout, port=dst_port,
-                                                  source=src_ip, source_port=src_port)
+                                                  source=src_ip, source_port=src_port,
+                                                  http_version=dns.query.HTTPVersion.HTTP_2)
+                    except dns.query.NoDOH:
+                        die("ERROR: python httpx module not available")
                     except httpx.ConnectError:
-                        print_stderr(f"The server did not respond to DoH on port {dst_port}", should_die=True)
+                        die(f"DoH connection failed on port {dst_port}")
                 else:
                     unsupported_feature("DNS-over-HTTPS (DoH)")
 
             elif proto is PROTO_HTTP3:
-                if hasattr(dns.query, 'quic'):
+                if hasattr(dns.query, 'https') and hasattr(dns.query, 'HTTPVersion') and hasattr(dns.query.HTTPVersion, 'H3'):
                     try:
                         # For hostnames, construct full URL; for IPs, use IP directly
                         if dnsserver_hostname != dnsserver_ip:
@@ -405,8 +463,7 @@ def main():
                                                   source=src_ip, source_port=src_port,
                                                   http_version=dns.query.HTTPVersion.H3)
                     except ConnectionRefusedError:
-                        print_stderr(f"The server did not respond to DNS-over-HTTPS/3 on port {dst_port}",
-                                     should_die=True)
+                        die(f"DoH3 connection refused on port {dst_port}")
                 else:
                     unsupported_feature("DNS-over-HTTPS/3 (DoH3)")
 
@@ -419,19 +476,18 @@ def main():
                                                  source=src_ip, source_port=src_port,
                                                  server_hostname=server_hostname)
                     except dns.exception.Timeout:
-                        print_stderr(f"The server did not respond to DoQ on port {dst_port}", should_die=True)
+                        die(f"DoQ connection timeout on port {dst_port}")
                     except ConnectionRefusedError:
-                        print_stderr(f"The server did not respond to DNS-over-HTTPS/3 on port {dst_port}",
-                                     should_die=True)
+                        die(f"DoQ connection refused on port {dst_port}")
                 else:
                     unsupported_feature("DNS-over-QUIC (DoQ)")
 
             etime = time.perf_counter()
         except dns.resolver.NoNameservers as e:
             if not quiet:
-                print_stderr("No response to DNS request", False)
+                err("No response to DNS request")
                 if verbose:
-                    print_stderr("error: %s" % e, False)
+                    err(f"ERROR: {e}")
             sys.exit(1)
         except (httpx.ConnectTimeout, dns.exception.Timeout):
             if not quiet:
@@ -439,17 +495,26 @@ def main():
         except httpx.ReadTimeout:
             if not quiet:
                 print("Read timeout", flush=True)
+        except EOFError:
+            if not quiet:
+                err("Connection closed by server")
         except PermissionError:
             if not quiet:
-                print_stderr("Permission denied", True)
-            sys.exit(1)
+                die("ERROR: permission denied")
+            else:
+                sys.exit(1)
         except OSError as e:
-            if not quiet:
-                print_stderr("%s" % e, True)
-            sys.exit(1)
+            if e.errno == 65:  # EHOSTUNREACH
+                die("ERROR: No route to host")
+            elif e.errno == 51:  # ENETUNREACH
+                die("ERROR: Network unreachable")
+            elif not quiet:
+                die(f"ERROR: {e}")
+            else:
+                sys.exit(1)
         except ValueError:
             if not quiet:
-                print_stderr("Invalid Response", False)
+                err("ERROR: invalid response")
                 continue
         except KeyboardInterrupt:
             # Handle Ctrl+C during DNS query
@@ -458,7 +523,8 @@ def main():
         else:
             # convert time to milliseconds, considering that
             # time property is returned differently by query.https
-            if type(answers.time) is datetime.timedelta:
+            # dns library returns float for most protocols but timedelta for HTTPS
+            if isinstance(answers.time, datetime.timedelta):
                 elapsed = answers.time.total_seconds() * 1000
             else:
                 elapsed = answers.time * 1000
@@ -510,10 +576,11 @@ def main():
                                 edns_parts.append("ECS:auto")
 
                 # Always show cookies when present (echoed back from server)
-                if answers.options:
+                if show_cookie and answers.options:
                     for ans_opt in answers.options:
                         if ans_opt.otype == 10:  # COOKIE
-                            edns_parts.append("COOKIE:%d" % len(ans_opt.data))
+                            cookie_hex = ans_opt.client.hex() + ans_opt.server.hex()
+                            edns_parts.append("COOKIE:%s" % cookie_hex)
 
                 if edns_parts:
                     extras += " [%s]" % ", ".join(edns_parts)
@@ -551,7 +618,9 @@ def main():
                             option_details = "code=%d, text=\"%s\"" % (ans_opt.code, ans_opt.text or "")
                         elif ans_opt.otype == 10:  # COOKIE
                             option_name = "COOKIE"
-                            option_details = "length=%d" % len(ans_opt.data)
+                            client_hex = ans_opt.client.hex()
+                            server_hex = ans_opt.server.hex()
+                            option_details = "client=%s, server=%s" % (client_hex, server_hex)
                         elif ans_opt.otype == 11:  # TCP-KEEPALIVE
                             option_name = "TCP-KEEPALIVE"
                             if len(ans_opt.data) >= 2:
@@ -571,7 +640,7 @@ def main():
                                 key_tags = []
                                 for i in range(0, len(ans_opt.data), 2):
                                     if i + 1 < len(ans_opt.data):
-                                        tag = struct.unpack('!H', ans_opt.data[i:i+2])[0]
+                                        tag = struct.unpack('!H', ans_opt.data[i:i + 2])[0]
                                         key_tags.append(str(tag))
                                 option_details = "tags=[%s]" % ",".join(key_tags)
                         else:
